// tabby  2005 Alexander Hajnal
// software@alephnull.net

// Compile with:
// gcc -o tabby -lXext  -lXtst -lX11 -lm -L/usr/X11R6/lib tabby.c

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#include <string.h>

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <X11/extensions/XTest.h>

// === Globals ===

// Display stucture for interfacing with X-Windows
Display    *display = NULL;
// Input file descriptor
int in;

// Since there is not a lot of precision in the proximity data we use a 
// running average to improve the resolution of the proximity sensor.   
// Unfortunately, the data is only accurate down to about a centimeter so 
// we use some tricks later (such as a weighted average) to reduce 
// positional jumps when the stylus is raised and lowered.
//
// PROX_BUF_SIZE: The size of the ring buffer.   Must be a multiple of 2 
//                because each set of data contains 2 proximity readings.
//                A large sized ring buffer provides greater the accuracy at 
//                a cost of reduced responsiveness.  Too low a value can lead 
//                to false stylus up and stylus down events.
// PROX_DOWN:     The proximity level (between 0 and 7) at which the stylus is 
//                considered to be down. Values between 4 and 4.5 are sensible
// PROX_UP:       The proximity level (between 0 and 7) at which the stylus is 
//                considered to be up. Values between 3 and 4 are sensible
#define PROX_BUF_SIZE 32
#define PROX_DOWN 4.3
#define PROX_UP 4.0
// The X/Y position also tends to be noisy.  The is a lot of low-level jitter 
// (1 or 2 pixels) even when the stylus is not being moved.  Occassionaly, a
// large jump (50 or more pixels) is erroniously reported.  There are two 
// paramters that can help alleviate this:
// 
// MIN_DISTANCE:  Motions of less than this distance will be ignored.
//                The distance is the length of the motion vector measured in 
//                pixels.  For example, a move of 3 pixels to the right and 4 
//                up would have a distance of 5 (3^2 + 4^2 = 5^2).  The default 
//                value of 2.5 can handle a jitter of about 2 pixels in either 
//                direction.
// MAX_DISTANCE:  Motions of greater than this distance will be ignored.
//                Again, the distance is the length of the motion vector 
//                measured in pixels.  This parameter should be set higher than 
//                the largest motion seen in normal usage.   If it is set too 
//                low then the pointer will stop tracking the stylus when it is 
//                moved quickly.
#define MIN_DISTANCE 2.5
#define MAX_DISTANCE 50

int prox_buf[PROX_BUF_SIZE]; // Proximity ring buffer
int prox_buf_pos;            // Current position in ring buffer

typedef struct sample {
	int x;
	int y;
	int prox;
	int prox2;
	int state;
} sample;

// Signal handler
void signal_handler(int signal) {
  switch (signal) {
	case SIGINT:
		// Close display If it has been opened
		if (display != NULL) {
			XCloseDisplay(display);
		}
		exit(0); // Graceful exit
		break;
	default:
		// Unexpected signal
		printf("Unxpected signal %d, terminating", signal);
  	_exit(signal); // Go down hard
  }
}

void read_sample(struct sample* sample) {
	unsigned char buf[6]; // Input buffer
	unsigned char d;      // Buffer for current byte read
	unsigned char id;     // Position in input buffer
	
	while (1) {
		if (read(in, &d, 1)) { // Read a byte off the serial port
			id = d >> 5; // Determine this byte's ID number
			buf[id] = d; // Record byte in input buffer
			if (id == 5) { // Byte with ID 5 is the last byte in complete postion read
				// Het the two proximity readings
				sample->prox  = (buf[0] >> 1) & 0x07;
				sample->prox2 = (buf[1] >> 1) & 0x07;
				// Decode position and button state
				sample->x = ( (buf[0] & 0x01) << 10) | 
				            ( (buf[2] & 0x1f) << 5) | 
				            (buf[4] & 0x1f);
				sample->y = ( (buf[1] & 0x01) << 10) | 
				            ( (buf[3] & 0x1f) << 5) | 
				            (buf[5] & 0x1f);
				sample->state = (buf[0] >> 4) & 0x01;
				//printf("x: %i  y: %i  state: %i  prox: %i %i\n", sample->x, sample->y, sample->state, sample->prox, sample->prox2);
				return;
			}
		}
	}
}

float get_smoothed_proximity(struct sample* sample) {
	float smoothed_prox;         // Average of values in ring buffer
	// Store the two promity readings in the ring buffer
	prox_buf[prox_buf_pos++ % PROX_BUF_SIZE]=sample->prox;
	prox_buf[prox_buf_pos++ % PROX_BUF_SIZE]=sample->prox2;
	// Determine the smoothed proximity
	{

		int i;             // Loop counter
		int found = 0;     // Number of proximity values found
		smoothed_prox = 0; // Initilalize to fully up
		for (i=0; i<PROX_BUF_SIZE; i++) { // Loop through ring buffer
			if (prox_buf[i] != -1) {        // Only use valid values
				// Add value to total and adjust datum count
				smoothed_prox = smoothed_prox + prox_buf[i];
				found++;
			}
		}
		// Compute average proximity
		if (found>0) {
			smoothed_prox = smoothed_prox / found;
		} else {
			smoothed_prox = 0;
		}
	}
	if (prox_buf_pos >= PROX_BUF_SIZE) prox_buf_pos -= PROX_BUF_SIZE;
	return smoothed_prox;
}

int main(int argc, char** argv) {
	struct sample current;
	struct sample last;
	
	int dx, dy;           // Position delta along each axis
	float distance;       // Linear distance of positional delta
	
	int valid;            // Non-zero after two samples are read
	
	float smoothed_prox;         // Average of values in ring buffer
	
  // More datastuctures for interfacing with X-Windows
  int         screen;
  Window      window;
  int result;
	
	
	// Program takes a single argument (the serial port device)
	if ( argc != 2 || 
	     (strcmp(argv[1], "-h") == 0) ||
			 (strcmp(argv[1], "--help") == 0) 
	   ) {
		printf("Usage: tabby <serial port>\n");
		exit(1);
	}
	
	// Set up signal handlers
	(void)signal(SIGINT, signal_handler);
	
	// Initialize X-Windows connection
	display = XOpenDisplay(NULL);               // Open display specified in DISPLAY environment variable
  if (display == NULL) { 
		printf("Could not open display\n");
		exit(1);
	}
  screen = DefaultScreen(display);            // Get default screen for display
  window = RootWindow(display, screen);       // Get the screen's root window
	result = XTestGrabControl(display , True);  // Grab display (what is the diffenece between True and False for impervious?)
  if (result!=1) {
		printf("Could not grab display\n");
		exit(1);
	}
	
	// Initialize the serial port by calling 'stty -F PORT raw 9600'
	// This code works on Linux systems.  Other systems may need changes.
	{
		pid_t pid;
		if (pid=fork()) {
			waitpid(pid, NULL, 0);
		} else {
			(void)execl("/bin/stty", "-F", argv[1], "raw", "9600");
			exit(0);
		}
	}
	
	// Open the serial port
	in=open(argv[1], O_RDONLY);
	if (in == -1) {
		perror("Could not open serial port");
		exit(1);
	}
	
	// Clear the proximity ring buffer
	for (prox_buf_pos=0; prox_buf_pos<PROX_BUF_SIZE; prox_buf_pos++) prox_buf[prox_buf_pos] = -1;
	
	valid = 0;        // Data in input buffer is not valid
	last.state = -1;      // No previous button state
	current.prox = current.prox2 = 0; // Proximity is fully off the surface
	
	// Main loop
	while (1) {
		read_sample(&current);
		smoothed_prox = get_smoothed_proximity(&current);
		
		
		// If sylus is determined to be down
		if (smoothed_prox > PROX_DOWN) {
			
			// Need two points to determine a vector
			if (valid) { // Two points have been read
				
				// Determine position delta on each axis
		    //printf("dx = -1*(%i - %i)\n", current.x, last.x);
				//printf("dy = -1*(%i - %i)\n", current.y, last.y);
				dx = -1*(current.x-last.x);
				dy = -1*(current.y-last.y);
				
				// Determine striaght-line distance of delta
				distance = sqrt(dx*dx+dy*dy);
				//printf("distance = sqrt(%i ^2 + %i ^2)\n", dx, dy);
				//printf("distance: %f\n", distance);
				
				if (distance<MIN_DISTANCE || distance>MAX_DISTANCE) {
					// Ignore small movements (indistinguishable from noise) 
					// and large movements (usually a result of noise)
					dx=0;
					dy=0;
					// Ideally, the next like should not be present - we need it though 
					// because the cursor can get into a locked state if either current 
					// value gets large enough (e.g. > 2000).  The code should detect 
					// this condition and recover from it.
					last.x=current.x; last.y=current.y;
				} else {
					// Store the current position and state
					last.x=current.x; last.y=current.y;
				}
				
	      // Pass motion event to X
				//printf("XTestFakeRelativeMotionEvent (%i, %i, %i, %i)\n", display, dx, dy, 0);
				if (dx || dy) {
					XTestFakeRelativeMotionEvent (display, dx, dy, 0);
				}
				// If button has changed state...
				if ( current.state != last.state ) {
					if (current.state == 0) { // Stylus down
						XTestFakeButtonEvent ( display, 1, True, 0);
					} else {          // Stylus up
						XTestFakeButtonEvent ( display, 1, False, 0);
					}
				}
				// Flush event queue
				XFlush(display);
			} else { // Called for the first point read
				// Store the current position and state
				last.x=current.x; last.y=current.y;
				// Note that a valid position has been read
				valid = 1;
			}
		} else if (smoothed_prox < PROX_UP) { // Stylus is up
			if ( current.state == 0 ) { // Button was pressed
				// Mark button as released
				current.state = 1;
				// Send button release event
				XTestFakeButtonEvent ( display, 1, False, 0);
				// Flush event queue
				XFlush(display);
			}
		}
		// Store button and proximity state
		last.state=current.state;
		last.prox=current.prox;
		last.prox2=current.prox2;
	} // End main loop
}
